import argparse
import asyncio
import datetime
import json
import os
import re
import time
from pathlib import Path

import aiohttp
import jinja2
from aiohttp import web
from html_sanitizer import Sanitizer
from markupsafe import Markup

BASE_DIR = Path(__file__).parent
BASE_URL = 'https://api.group.heimstaden.com'

sanitizer = Sanitizer({
    'tags': {'p', 'a', 'ul', 'li', 'strong', 'em', 'br', 'hr', 'div'},
    'separate': {'a', 'p', 'li', 'div'},
})
env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(BASE_DIR / 'templates'),
    autoescape=jinja2.select_autoescape(default=True),
    trim_blocks=True,
    lstrip_blocks=True,
)


def timestamp2date(value):
    return datetime.datetime.fromtimestamp(value / 1000).date()
env.filters['timestamp2date'] = timestamp2date


def sanitize(html):
    return Markup(sanitizer.sanitize(html))
env.filters['sanitize'] = sanitize


def linebreaks(text):
    return Markup('').join(
        Markup('<p>{}</p>').format(
            Markup('<br>').join(p.split('\n'))
        )
        for p in re.split(r'\n\n+', text)
    )
env.filters['linebreaks'] = linebreaks


@jinja2.pass_context
def translate(context, key):
    lang = context['preferences']['language'].split('_')[0]
    for item in context['translations']:
        if item['key'] == key:
            return item[lang]
    raise KeyError(key)
env.filters['translate'] = translate


@jinja2.pass_context
def category(context, cat, field=None):
    lang = context['preferences']['language'].split('_')[0]
    if cat in context['categories']:
        item = context['categories'][cat]
    else:
        item = context['serviceRequestCategories'][cat]
    if field:
        return item[f'{lang}_{field}']
    else:
        return item[lang].replace('|', ' · ')
env.filters['category'] = category


@jinja2.pass_context
def ticket_status(context, status, field=None):
    colors = {
        '#CCDCFF': 'in-progress',
        '#CCEFD9': 'done',
        '#CCF6F1': 'waiting',
        '#ECEEF0': 'on-hold',
        '#F4D3EF': 'open',
    }
    for item in context['ticketStatuses']:
        if item['name'] == status:
            if field == 'class':
                return colors.get(item['colorHex'])
            else:
                return translate(context, item['translation'])
    raise KeyError(status)
env.filters['ticket_status'] = ticket_status


@jinja2.pass_context
def notification_body(context, notification):
    t = translate(context, notification['translation']['bodyKey'])
    t = t.removeprefix('rich-text:')
    if '{{title}}' in t:
        t = t.replace('{{title}}', notification['metadata']['ticketTitle'])
    if '{{status}}' in t:
        status = ticket_status(context, notification['metadata']['newStatus'])
        t = t.replace('{{status}}', status)
    return sanitize(t)
env.filters['notification_body'] = notification_body


@jinja2.pass_context
def ticket_events(context, ticket):
    events = [{
        'type': 'ticket',
        'created_at': ticket['created_at'],
        'data': ticket,
    }]
    for comment in ticket['comments']:
        events.append({
            'type': 'comment',
            'created_at': comment['created_at'],
            'data': comment,
        })
    for notification in context['notifications']:
        if (
            notification['metadata']['ticketId'] == str(ticket['id'])
            and notification['type'] not in ['ticket_commented', 'ticket_created']
        ):
            events.append({
                'type': 'notification',
                'created_at': notification['created_at'],
                'data': notification,
            })
    return sorted(events, key=lambda event: event['created_at'])
env.filters['ticket_events'] = ticket_events


async def fetch(session, *args, **kwargs):
    print('fetch', args, kwargs)
    async with session.request(*args, **kwargs) as response:
        response.raise_for_status()
        return await response.json()


async def fetch_paged(session, *args, params={}, **kwargs):
    items = []
    while True:
        page = await fetch(session, *args, params=params, **kwargs)
        items += page['items']
        if page['nextPage']:
            params = {**params, 'page': page['nextPage']}
        else:
            return items


async def fetch_unwrap_lease(session, *args, keep=False, **kwargs):
    data = await fetch(session, *args, **kwargs)
    assert len(data) == 1
    if keep:
        return data[0]
    else:
        assert len(data[0]) == 2
        for key, value in data[0].items():
            if key != 'lease_id':
                return value

async def maybe(awaitable, default):
    try:
        return await awaitable
    except Exception as e:
        print(e)
        return default


async def fetch_tickets(session):
    tickets = await fetch_paged(session, 'GET', '/api:Q30QZMaz/tickets', params={
        'per_page': 100, 'sort_by': '-created_at'
    })

    details = await asyncio.gather(*[
        asyncio.gather(
            fetch(session, 'GET', f'/api:Q30QZMaz/tickets/{ticket["id"]}'),
            maybe(
                # TODO: why does this not work for all tickets?
                fetch_paged(
                    session,
                    'GET',
                    f'/api:Q30QZMaz/tickets/{ticket["id"]}/comments',
                    params={
                        'order_by': 'asc',
                        'page': 1,
                        'per_page': 100,
                        'sort_by': 'created_at',
                    },
                ),
                [],
            ),
        ) for ticket in tickets
    ])

    return [{**ticket, 'comments': comments} for ticket, comments in details]


async def _get_token(session):
    async with session.post(
        '/api:stKpHQTZ/auth/login',
        json={
            'email': os.getenv('HEIMSTADEN_EMAIL'),
            'password': os.getenv('HEIMSTADEN_PASSWORD'),
        },
    ) as response:
        response.raise_for_status()
        return await response.json()


async def _get_data(session):
    data = {}
    (
        data['translations'],
        data['categories'],
        data['serviceRequestCategories'],
        data['ticketStatuses'],
        data['me'],
        data['preferences'],
        data['leases'],
        data['balance'],
        data['rent'],
        data['transaction'],
        data['payment'],
        data['bank_account'],
        data['oc_documents'],
        data['tickets'],
        data['notifications'],
        data['emails'],
        data['broadcastMessages'],
    ) = await asyncio.gather(
        fetch(session, 'GET', '/api:oJE6MLtZ/myhome_translations?exclude_categories'),
        fetch(session, 'POST', '/api:Q30QZMaz/categories', json={'catalog': 'my_account'}),  # noqa
        fetch(session, 'POST', '/api:Q30QZMaz/categories', json={'catalog': 'service_request'}),  # noqa
        fetch(session, 'GET', '/api:Q30QZMaz/ticket-statuses'),
        fetch(session, 'GET', '/api:Q30QZMaz/me'),
        fetch(session, 'GET', '/api:Q30QZMaz/preferences'),
        fetch_paged(session, 'GET', '/api:Q30QZMaz/leases'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/finance/balance'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/finance/rent'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/finance/transaction'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/finance/payment'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/bank_account'),
        fetch_unwrap_lease(session, 'GET', '/api:Q30QZMaz/oc_documents', keep=True),
        fetch_tickets(session),
        fetch_paged(session, 'GET', '/api:Q30QZMaz/notifications', params={
            'page_size': 100
        }),
        fetch(session, 'GET', '/api:Q30QZMaz/emails'),
        fetch(session, 'GET', '/api:Q30QZMaz/broadcast/messages'),
    )

    assert len(data['leases']) == 1
    data['lease'] = data['leases'][0]
    del data['leases']

    return data


async def get_data():
    dump = BASE_DIR / 'dump.json'

    if dump.exists():
        with dump.open() as fh:
            data = json.load(fh)
        if data.get('time', 0) + 60 * 60 >= time.time():
            return data

    async with aiohttp.ClientSession(base_url=BASE_URL) as session:
        token = await _get_token(session)
        session.headers['Authorization'] = f'Bearer {token["authToken"]}'
        data = await _get_data(session)
    data['time'] = time.time()

    with dump.open('w') as fh:
        json.dump(data, fh, indent=2)

    return data


async def view(request):
    data = await get_data()
    tpl = env.get_template('index.html')
    body = tpl.render(**data)
    return web.Response(body=body, content_type='text/html')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=8000)
    args = parser.parse_args()

    app = web.Application(middlewares=[
        web.normalize_path_middleware(),
    ])
    app.router.add_static('/static', BASE_DIR / 'static')
    app.router.add_get('/', view)
    web.run_app(app, host='localhost', port=args.port)
